import time
import types

import msgpack


class BaseServer(object):
    SOCKET_BUFFER_SIZE = 1024
    ALLOWED_TYPES = (types.FunctionType, types.MethodType, types.BuiltinFunctionType, types.BuiltinMethodType)

    def __init__(self, services):
        self.services = self.load_services(services)
        if not self.services:
            raise TypeError("First argument to BaseServer.__init__ must be a dict or a string")

        self.packer = msgpack.Packer()

    def load_services(self, module):
        services = {}
        if isinstance(module, dict):
            services.update(module)
        elif isinstance(module, types.ModuleType):
            for name in dir(module):
                if not name.startswith("_"):
                    attr = getattr(module, name)
                    if isinstance(attr, self.ALLOWED_TYPES):
                        services[name] = attr
        elif isinstance(module, basestring):
            services.update(self.load_services(__import__(module)))
        elif isinstance(module, (tuple, list)):
            for m in module:
                services.update(self.load_services(m))
        return services

    def unpack_requests(self, sock):
        unpacker = msgpack.Unpacker()
        while True:
            data = sock.recv(self.SOCKET_BUFFER_SIZE)
            if not data:
                break
            unpacker.feed(data)
            for request in unpacker:
                yield request

    def pack_result(self, result):
        return self.packer.pack(result)

    def handle_request(self, request):
        start_time = time.time()
        req_method = request.get("method")
        req_args = request.get("args", [])

        result = {
            "id": request.get("id"),
            "results": None,
        }

        if not req_method:
            result["error"] = "No request method was provided"
        elif not isinstance(self.services.get(req_method), self.ALLOWED_TYPES):
            result["error"] = "Unknown request method '%s'" % (req_method, )
        else:
            try:
                result["results"] = self.services[req_method](*req_args)
            except Exception, e:
                result["error"] = e.message

        result["run_time"] = (time.time() - start_time) * 1000.0
        return result


class BaseClient(object):
    SOCKET_BUFFER_SIZE = 1024

    def __init__(self):
        self.unpacker = msgpack.Unpacker()
        self.packer = msgpack.Packer()
        self.id = 0

    def pack_request(self, id, method, args=[]):
        return self.packer.pack({
            "id": id,
            "method": method,
            "args": args,
        })

    def unpack_results(self, sock):
        while True:
            data = sock.recv(self.SOCKET_BUFFER_SIZE)
            if not data:
                break
            self.unpacker.feed(data)
            return self.unpacker.next()

    def __getattr__(self, method):
        def wrapper(*args):
            return self.call(method, args)
        return wrapper



import contextlib
import json
import urlparse

from gevent import pywsgi
from gevent.server import StreamServer
import msgpack

from greenrpc import DEFAULT_PORT
from greenrpc.base import BaseServer


class TCPServer(StreamServer, BaseServer):
    def __init__(self, services, bind=("127.0.0.1", DEFAULT_PORT), spawn=1):
        StreamServer.__init__(self, bind, spawn=spawn)
        BaseServer.__init__(self, services)

    def handle(self, socket, address):
        with contextlib.closing(socket) as sock:
            with contextlib.closing(sock.makefile()) as fp:
                for request in self.unpack_requests(sock):
                    result = self.handle_request(request)
                    fp.write(self.pack_result(result))
                    fp.flush()


class WSGIServer(pywsgi.WSGIServer, BaseServer):
    def __init__(self, services, bind=("127.0.0.1", DEFAULT_PORT), spawn=1):
        pywsgi.WSGIServer.__init__(self, bind, spawn=spawn)
        BaseServer.__init__(self, services)

    def decode(self, decoder, data):
        decoder = decoder.lower()
        if decoder == "json":
            return json.loads(data)
        elif decoder == "msgpack":
            return msgpack.unpackb(data)

    def encode(self, encoder, data):
        encoder = encoder.lower()
        if encoder == "json":
            return json.dumps(data)
        elif encoder == "msgpack":
            return msgpack.packb(data)

    def content_type(self, encoder):
        encoder = encoder.lower()
        mime = "text/plain"
        if encoder == "json":
            mime = "application/json"
        elif encoder == "msgpack":
            mime = "application/x-msgpack"
        return ("Content-Type", mime)

    def application(self, environ, start_response):
        content_length = int(environ.get("CONTENT_LENGTH", 0))
        uri = environ["PATH_INFO"]
        uri = uri.strip("/")
        encoder, _, method = uri.partition("/")

        qs = urlparse.parse_qs(environ["QUERY_STRING"])
        request = {
            "id": qs.get("id", [None])[0],
            "args": qs.get("arg", []),
            "method": method,
        }

        if environ["REQUEST_METHOD"] == "POST":
            data = environ["wsgi.input"].read(content_length)
            data = self.decode(encoder, data)
            request["method"] = data.get("method", request["method"])
            request["args"] = data.get("args", request["args"])
            request["id"] = data.get("id", request["id"])

        result = self.handle_request(request)
        start_response("200 OK", [self.content_type(encoder)])
        return [self.encode(encoder, result)]



from gevent.socket import socket
import msgpack
import requests

from greenrpc import DEFAULT_PORT
from greenrpc.base import BaseClient
from greenrpc.error import RPCException


class TCPClient(BaseClient):
    def __init__(self, connect=("127.0.0.1", DEFAULT_PORT)):
        super(TCPClient, self).__init__()
        self.connection = socket()
        self.connection.connect(connect)
        self.fp = self.connection.makefile()

    def call(self, method, args=[], debug=False):
        self.id += 1
        request = self.pack_request(self.id, method, args)
        self.fp.write(request)
        self.fp.flush()
        results = self.unpack_results(self.connection)
        if debug:
            return results
        if results.get("error"):
            raise RPCException(results["error"])
        return results.get("results")


class HTTPClient(BaseClient):
    def __init__(self, connect=("127.0.0.1", DEFAULT_PORT)):
        super(HTTPClient, self).__init__()
        self.base_url = "http://%s:%s/msgpack" % connect

    def call(self, method, args=[], debug=False):
        self.id += 1
        request = self.pack_request(self.id, method, args)
        result = requests.post(self.base_url, request)
        data = msgpack.unpackb(result.content)
        if debug:
            return data
        if data.get("error"):
            raise RPCException(data["error"])
        return data.get("results")



#!/usr/bin/env python

from setuptools import setup, find_packages

from greenrpc import __version__

with open("./requirements.txt") as fp:
    requirements = fp.read()
    requirements = requirements.split("\n")

setup(
    name="greenrpc",
    version=__version__,
    description="TCP & HTTP RPC Servers written with msgpack and gevent",
    author="Brett Langdon",
    author_email="brett@blangdon.com",
    url="https://github.com/brettlangdon/greenrpc",
    packages=find_packages(),
    license="MIT",
    scripts=["bin/greenrpc-server", "bin/greenrpc-client"],
    install_requires=requirements,
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "License :: OSI Approved :: MIT License",
        "Topic :: Utilities",
    ]
)
